*:*********************************************************************
*:
*:        Program: BROWSE.PRG
*:
*:         System: The TBrowse from HELL!
*:         Author: E. Mark Kothe
*:      Copyright (c) 1991, E. Mark othe
*:  Last modified: 07/29/91     15:58
*:
*:  Procs & Fncts: MYBROWSE()
*:               : SKIPWHILE()
*:               : DOGET()
*:               : POPSEEK()
*:               : KEYSEEK()
*:               : TBWHILESET()
*:               : GETTOP()
*:               : GETBOT()
*:               : SEEKLAST()
*:
*:      Documented 07/29/91 at 15:58                SNAP!  version 4.02d
*:*********************************************************************
*!*********************************************************************
*!
*!       Function: MYBROWSE()
*!
*!          Calls: TBWHILESET()   (function  in BROWSE.PRG)
*!               : SKIPWHILE()    (function  in BROWSE.PRG)
*!               : GETTOP()       (function  in BROWSE.PRG)
*!               : GETBOT()       (function  in BROWSE.PRG)
*!               : KEYSEEK()      (function  in BROWSE.PRG)
*!               : POPSEEK()      (function  in BROWSE.PRG)
*!               : DOGET()        (function  in BROWSE.PRG)
*!
*!*********************************************************************

* Sets Up nTopNo And nBotNo For TBWhile()
static nbotno,ntopno
* the old seek key
static cSeekkey := ''

FUNCTION mybrowse (ntop, nleft, nbottom, nright,;
   lkeyseek,cfields, lallow,;
   lwantrecno, lwanthead, nfreeze, ckey)


* v1.1 fixed to actually work, Sorry! - E. Mark Kothe
* v1.2 added while condition processing
*      NOW called the TBrowse from HELL!
* v1.3 added sending a logical for headings or not, or
*      an array of heading names
*
*       MyBrowse.prg
*       Illustration of TBROWSE and GET objects
*
* Calling parameters are:
*             nTop - top row
*            nLeft - leftmost column
*          nBottom - bottom row
*           nRight - right column
*         lKeyseek - do keyseek searches, just by typing?
*          cFields - a string of field numbers to display
*           lAllow - allow editing?
*       lWantrecno - Want a record number?
*        lWanthead - It's not what you think, it' do you want a header?
*                    This parameter can be a logical, or an array or
*                    heading names
*          nFreeze - the number of columns to freeze
*             cKey - TRIMMED character string filter expression
*                    (an ALLTRIM is a must unless you pass an exact copy
*                     of the index key!)

#include "inkey.ch"
#include "setcurs.ch"

* column seperators
#define my_hsep         ""
#define my_csep         "  "

local B, column, ctype, N, bWhileCond
local ccolorsave, ncurssave
local lmore, nkey, lappend, where

* clear for this search
cSeekkey := ''

IF EMPTY(ALIAS())
   * no database in use
   RETURN .F.
ENDIF

* set up defaults

IF nbottom = nil
   nbottom=MAXROW()
ENDIF

IF ntop = nil
   ntop=1
ENDIF

IF nleft = nil
   nleft=0
ENDIF

IF nright = nil
   nright=Maxcol()
ENDIF

IF ntop<1
   ntop=1
ENDIF

IF lkeyseek = nil
   lkeyseek=.F.
ENDIF

* alow them to edit?
IF lallow = nil
   lallow=.T.
ENDIF

* did they pass a string of field numbers
IF cfields=nil
   cfields:=''
ENDIF

IF lwantrecno=nil
   lwantrecno:=.T.
ENDIF

IF lwanthead=nil
   lwanthead:=.T.
ENDIF

IF nfreeze=nil
   IF (lwantrecno,nfreeze := 1,nfreeze := 0)
   ENDIF
   
   IF ckey=nil
      ckey:=''
   ENDIF
   
   IF .NOT. tbwhileset(ckey)  //  passes KEY TO SET ntopno and nbotno
      RETURN(.F.)
   ENDIF
   
   * make new browse object call 'B'
   B := TBROWSEDB(ntop, nleft, nbottom, nright)
   
   * default heading and column separators for 'B' object
   IF lwanthead
      B:headsep := my_hsep
      B:colsep := my_csep
   ELSE
      B:headsep := '   '
      B:colsep := '   '
   ENDIF
   /* ADD custom 'SkipWhile' (TO handle passed condition & APPEND) */
   IF EMPTY(Indexkey(0))
      bwhilecond:= {||.t.}
   ELSE
      bwhilecond:={||&(Indexkey(0)) = ckey}
   ENDIF
   B:skipblock := {|x| skipwhile(X,bwhilecond ,lappend)}
   
   /* SET UP substitute GOTO TOP and GOTO BOTTOM */
   /* WITH WHILE's top and bottom records        */
   B:gotopblock := {|| DBGOTO(gettop())}
   B:gobottomblock := {|| DBGOTO(getbot())}
   
   /* ADD A column FOR RECNO() */
   IF lwantrecno
      column := Tbcolumnnew( "  Rec #", {|| RECNO()} )
      B:ADDCOLUMN(column)
   ENDIF
   
   IF EMPTY(cfields)
      
      * add a column for each field in the current workarea
      FOR N = 1 TO FCOUNT()

         * make the new column object
         * is it a logical or an array?
         IF valtype(lwanthead) = 'L'
           IF lWanthead
             column := Tbcolumnnew( Fieldname(N), Fieldblock(Fieldname(N)) )
           ELSE
             column := Tbcolumnnew( '', Fieldblock(Fieldname(N)) )
           ENDIF
         ELSE
           * is the array shorter than the # fields
           if N > LEN(lWanthead)
             column := Tbcolumnnew( Fieldname(N), Fieldblock(Fieldname(N)) )
           ELSE
             column := Tbcolumnnew( lWanthead[N], Fieldblock(Fieldname(N)) )
           endif
         ENDIF
         
         * add the column object to the Browse object
         B:ADDCOLUMN(column)
      NEXT n
      
   ELSE
      
      * v1.1 Trim all outer spaces
      cfields:=Alltrim(cfields)
      
      * add a column for each REQUESTED field in the current workarea
      * fields are in a string such as '1 12 3 4 5 21'
      DO WHILE ! EMPTY(cfields)
         
         N:=VAL(cfields)
         
         * make the new column object
         * is it a logical or an array v2.1?
         IF valtype(lwanthead) = 'L'
           IF lWanthead
             column := Tbcolumnnew( Fieldname(N), Fieldblock(Fieldname(N)) )
           ELSE
             column := Tbcolumnnew( '', Fieldblock(Fieldname(N)) )
           ENDIF
         ELSE
           * is the array shorter than the # fields
           if N > LEN(lWanthead)
             column := Tbcolumnnew( Fieldname(N), Fieldblock(Fieldname(N)) )
           ELSE
             column := Tbcolumnnew( lWanthead[N], Fieldblock(Fieldname(N)) )
           endif
         ENDIF
         
         * add the column object to the Browse object
         B:ADDCOLUMN(column)
         
         * v1.1 Sorry, gotta check if we got the end of string
         where:=AT(' ',cfields)
         IF where != 0
            cfields := SUBSTR(cfields,where+1)
         ELSE
            EXIT
         ENDIF
         
      ENDDO
   ENDIF
   
   * freeze leftmost column(s)
   B:FREEZE := nfreeze
   
   * save the cursor status
   ncurssave := Setcursor(0)
   
   * not in append mode
   lappend := .F.
   
   * a jump start
   lmore := .T.
   DO WHILE (lmore)
      
      * don't allow cursor to move into frozen columns
      IF ( B:colpos <= B:FREEZE )
         B:colpos := B:FREEZE + 1
      ENDIF
      
      * stabilize the display before we wait
      DO WHILE ( ! B:Stabilize() )
         nkey := INKEY()
         IF ( nkey != 0 )
            * abort if a key is waiting
            EXIT
         ENDIF
      ENDDO
      
      * we got a key, assure the screen is stable!
      IF ( B:stable )
         
         * we are at the bottom and not in append mode
         IF ( B:hitbottom .AND. ! lappend )
            * banged against EOF; go into append mode
            lappend := .T.
            nkey := K_DOWN
         ELSE
            IF ( B:hittop .OR. B:hitbottom )
               * grunt at em'
               TONE(125, 0)
            ENDIF
            /* everything's done; just wait for a key */
            nkey := INKEY(0)
         ENDIF
      ENDIF
      
      * keyboard processor
      DO CASE
      CASE ( nkey == K_DOWN )
         B:DOWN()
         
      CASE ( nkey == K_UP )
         B:UP()
         IF ( lappend )
            lappend := .F.
            B:Refreshall()
         ENDIF
         
      CASE ( nkey == K_PGDN )
         B:Pagedown()
         
      CASE ( nkey == K_PGUP )
         B:Pageup()
         IF ( lappend )
            lappend := .F.
            B:Refreshall()
         ENDIF
         
      CASE ( nkey == K_CTRL_PGUP )
         B:Gotop()
         lappend := .F.
         
      CASE ( nkey == K_CTRL_PGDN )
         B:Gobottom()
         lappend := .F.
         
      CASE ( nkey == K_RIGHT )
         B:RIGHT()
         
      CASE ( nkey == K_LEFT )
         B:LEFT()
         
      CASE ( nkey == K_HOME )
         B:Home()
         
      CASE ( nkey == K_END )
         B:END()
         
      CASE ( nkey == K_CTRL_LEFT )
         B:Panleft()
         
      CASE ( nkey == K_CTRL_RIGHT )
         B:Panright()
         
      CASE ( nkey == K_CTRL_HOME )
         B:Panhome()
         
      CASE ( nkey == K_CTRL_END )
         B:Panend()
         
      CASE ( nkey == K_ESC )
         lmore := .F.
         
         * on backspace clear the key search string
      CASE ( nkey == K_BS)
         keyseek(.T.)
         
      CASE ( nkey == K_ALT_F2 )
         *seek mode on alt f2
         IF popseek()
            B:Refreshall()
         ENDIF
         
      CASE ( nkey == K_RETURN )
         * do we allow them to edit?
         IF lallow
            doget(B, lappend)
            tbwhileset(ckey)
            B:Refreshall()
         ELSE
            lmore:=.F.
         ENDIF
         
      OTHERWISE
         * don't know what this key is, either
         * seek for it, or do a get or abandon!
         IF lkeyseek
            keyseek(CHR(nkey))
            B:Refreshall()
         ELSE
            IF lallow
               KEYBOARD( CHR(nkey) )
               doget(B, lappend)
               tbwhileset(ckey)
               B:Refreshall()
            ELSE
               lmore := .F.
            ENDIF
         ENDIF
         
      ENDCASE
   ENDDO
   
   * restore cursor
   Setcursor(ncurssave)
   
   RETURN (.T.)
   
   
   
*!*********************************************************************
*!
*!       Function: SKIPWHILE()
*!
*!      Called by: MYBROWSE()     (function  in BROWSE.PRG)
*!
*!*********************************************************************
   FUNCTION skipwhile (N, bwhilecond, lappend)
   *
   *       Skipper()
   *
   * parameters are:
   *           n - number of rows to skip, positive or negative
   *     lAppend - append status
   *
   local I
   
   I := 0
   
   * empty DBF
   IF ( LASTREC() != 0 )
      IF ( N == 0 )
         * just refresh data
         SKIP 0
         
         * going down, and not at end of file!
      ELSEIF ( N > 0 .AND. RECNO() != LASTREC() + 1 )
         DO WHILE ( I < N )
            SKIP 1
            IF ( EOF() )
               IF ( lappend )
                  I++
               ELSE
                  SKIP -1
               END
               EXIT
            ELSE
               IF .NOT. EVAL(bwhilecond)
                  SKIP -1
                  EXIT
               ENDIF
               
            ENDIF
            I++
         ENDDO
         * skipping up
      ELSEIF ( N < 0 )
         DO WHILE ( I > N )
            SKIP -1
            * careful 5.0 bof is FUSSY, DO not
            * test any data, or BOF() will go
            * false, every time!  Just another
            * of those intended behaviors?
            IF ( BOF() )
               EXIT
            ELSE
               IF .NOT. EVAL( (bwhilecond) )
                  SKIP
                  EXIT
               ENDIF
            ENDIF
            I--
         ENDDO
      ENDIF
   ENDIF
   
   * tell them how far we went!
   RETURN (I)
   
   *
*!*********************************************************************
*!
*!       Function: DOGET()
*!
*!      Called by: MYBROWSE()     (function  in BROWSE.PRG)
*!
*!*********************************************************************
   FUNCTION doget (B, lappend)
   *
   *       DoGet()
   *
   * parameters are:
   *        b - the browse Object
   *  lAppend - the append status
   *
   local binssave, lscoresave, lexitsave
   local column, GET, nkey
   
   * make sure browse is stable
   DO WHILE ( !B:Stabilize() ) ; END
      
      
      * if confirming new record, append blank
      IF ( lappend .AND. RECNO() == LASTREC() + 1 )
         * this shoud be a ADD_REC(5), network aware
         * append !
         APPEND BLANK
      ENDIF
      
      * Sorry, this is my function for locking a record
      * substitute your own, or beg me for the source
      * on mine!
      IF ! REC_LOCK(5)
         * another of mine to just display a message and wait
         YESNO('Cannot lock this record',3)
         RETURN(nil)
      ENDIF
      
      
      * save state
      lscoresave := SET(_SET_SCOREBOARD, .F.)
      lexitsave := SET(_SET_EXIT, .T.)
      binssave := SETKEY(K_INS)
      
      * set insert key to toggle insert mode and cursor
      SETKEY( K_INS, ;
         {|| Setcursor( IF(Readinsert(!Readinsert()), SC_NORMAL, SC_INSERT))})
      
      * initial cursor setting
      Setcursor( IF(Readinsert(), SC_INSERT, SC_NORMAL) )
      
      /* GET column object FROM BROWSE */
      column := B:GETCOLUMN(B:colpos)
      
      /* CREATE A corresponding GET object*/
      GET := GETNEW(ROW(), COL(), column:block, column:HEADING,, B:colorspec)
      
      * read it */
      READMODAL( {GET} )
      
      /* RESTORE state */
      Setcursor(0)
      SET(_SET_SCOREBOARD, LSCORESAVE)
      SET(_SET_EXIT, LEXITSAVE)
      SETKEY(K_INS, binssave)
      
      /* force redisplay of current ROW */
      B:REFRESHCURRENT()
      
      /* check EXIT KEY */
      nkey := LASTKEY()
      IF ( nkey == K_UP .OR. nkey == K_DOWN .OR. ;
            nkey == K_PGUP .OR. nkey == K_PGDN )
         KEYBOARD( CHR(nkey) )
      ENDIF
      UNLOCK
      RETURN (nil)
      
      
*!*********************************************************************
*!
*!       Function: POPSEEK()
*!
*!      Called by: MYBROWSE()     (function  in BROWSE.PRG)
*!
*!*********************************************************************
      FUNCTION popseek
      
      * a popup up seek function, crude but effective!
      
      local TEMP,getvar,idxkey,getlen,sgetvar,currec,getlist
      getlist := {}
      * get the index key
      idxkey=Indexkey(0)
      
      * is there an index?
      IF ! EMPTY(idxkey)
         currec=RECNO()
         SET CURSOR ON
         STORE &idxkey TO getvar
         SAVE SCREEN TO TEMP
         DO WHILE .T.
            IF VALTYPE(getvar)="C"
               IF LEN(getvar)>77
                  getlen=77
               ELSE
                  getlen=LEN(getvar)
               ENDIF
            ELSE
               getlen=10
            ENDIF
            IF getlen < LEN('Seeking on '+ UPPER(idxkey)+', Enter search string')
               getlen=LEN('Seeking on '+ UPPER(idxkey)+', Enter search string')
            ENDIF
            @1,77-getlen CLEAR TO 5,79
            @1,77-getlen TO 5,79 DOUBLE
            @2,78-getlen SAY 'Seeking on '+UPPER(idxkey)+', Enter search string'
            IF getlen=77
               @3,78-getlen GET getvar PICTURE "@KS50"
            ELSE
               @3,78-getlen GET getvar PICTURE "@K"
            ENDIF
            READ
            IF LASTKEY()=27
               RESTORE SCREEN FROM TEMP
               SET CURSOR OFF
               GOTO currec
               RETURN(.F.)
            ENDIF
            IF VALTYPE(getvar)="C"
               sgetvar=UPPER(RTRIM(getvar))
            ELSE
               sgetvar=getvar
            ENDIF
            SEEK sgetvar
            IF .NOT. FOUND()
               INKEY(1)
               IF VALTYPE(getvar)="C"
                  sgetvar=RTRIM(NOPUNC(sgetvar))
                  SEEK sgetvar
                  IF .NOT. FOUND()
                     @4,78-getlen SAY '**NOT Found**'
                     INKEY(1)
                     LOOP
                  ELSE
                     RESTORE SCREEN FROM TEMP
                     SET CURSOR OFF
                     RETURN(.T.)
                  ENDIF
               ELSE
                  @4,78-getlen SAY '**NOT Found**'
                  INKEY(1)
                  LOOP
               ENDIF
            ELSE
               RESTORE SCREEN FROM TEMP
               SET CURSOR OFF
               RETURN(.T.)
            ENDIF
         ENDDO
      ELSE
         YESNO("Can't SEEK a non indexed file",3)
         RETURN(.F.)
      ENDIF
      RETURN(nil)
      
      
*!*********************************************************************
*!
*!       Function: KEYSEEK()
*!
*!      Called by: MYBROWSE()     (function  in BROWSE.PRG)
*!
*!*********************************************************************
      FUNCTION keyseek (ckey)
      *
      * Seek while typing
      *
      
      * request to clear the seek key
      IF VALTYPE(ckey) = 'L'
         @0,0 CLEAR TO 0,LEN(cseekkey)
         cseekkey=''
         RETURN(.T.)
      ENDIF
      
      * what it was, plus last typed
      SEEK cseekkey+ckey
      
      * well that failed, try our special index type
      IF ! FOUND()
         SEEK Alltrim(NOPUNC(cseekkey+ckey))
      ENDIF
      
      * get it yet?
      IF FOUND()
         * show what we found
         cseekkey+=ckey
         @0,0 SAY cseekkey
         RETURN(.T.)
      ELSE
         * go back where we were, befor failing
         SEEK cseekkey
         TONE(100,3)
         RETURN(.F.)
      ENDIF
      RETURN(nil)
      
*!*********************************************************************
*!
*!       Function: TBWHILESET()
*!
*!      Called by: MYBROWSE()     (function  in BROWSE.PRG)
*!
*!          Calls: SEEKLAST()     (function  in BROWSE.PRG)
*!
*!*********************************************************************
      FUNCTION tbwhileset (ckey)
      local nCurrec
      IF EMPTY(ckey)  // IF no ckey, SET UP FOR ALL records
         nCurrec:=recno()
         GOTO BOTTOM
         nbotno := RECNO()
         GOTO TOP
         ntopno := RECNO()
         * return to where it was for non indexed
         goto nCurrec
         RETURN(.T.)
      ENDIF
      
      SEEK ckey
      IF EOF() .OR. LASTREC() == 0
         ntopno := nbotno := 0
         RETURN(.F.)
      ENDIF
      
      ntopno := RECNO()
      * goto last record for cKey in indexed dbf
      seeklast(ckey)
      * don't need to check if keyfield = cKey since previous
      * Seek cKey/IF EOF()/Return(.f.) already showed data present
      nbotno := RECNO()
      GOTO ntopno
      RETURN(.T.)
      * EOFcn TBWhileSet
      
*!*********************************************************************
*!
*!       Function: GETTOP()
*!
*!      Called by: MYBROWSE()     (function  in BROWSE.PRG)
*!
*!*********************************************************************
      FUNCTION gettop
      RETURN(ntopno)
      
*!*********************************************************************
*!
*!       Function: GETBOT()
*!
*!      Called by: MYBROWSE()     (function  in BROWSE.PRG)
*!
*!*********************************************************************
      FUNCTION getbot
      RETURN(nbotno)
      
*!*********************************************************************
*!
*!       Function: SEEKLAST()
*!
*!      Called by: TBWHILESET()   (function  in BROWSE.PRG)
*!
*!*********************************************************************
      FUNCTION seeklast (ckey)
      local lOldseek
      * SeekLast: Finds Last Record For Matching Key
      * Developed By Jon Cole
      * With softseek set on, seek the first record after condition.
      * This is accomplished by incrementing the right most character of the
      * string cKey by one ascii character.  After SEEKing the new string,
      * back up one record to get to the last record which matches cKey.
      lOldseek:=SET(_SET_SOFTSEEK,.t.)
      SEEK LEFT(ckey, LEN(ckey) -1) + CHR( ASC( RIGHT(ckey,1) ) +1)
      SET(_SET_SOFTSEEK,lOldseek)
      SKIP -1
      RETURN nil
      * EOFcn SeekLast
      
*: EOF: BROWSE.PRG
